/*
*
* Author: Philipp Zschoche, https://zschoche.org
*
*/
#ifndef __BLOSSOM5_HPP__
#define __BLOSSOM5_HPP__

#include "../blossom5/PerfectMatching.h"
#include "logger.hpp"
#include <vector>
#include <algorithm>
#include <chrono>


namespace mmatching {

  struct blossom_result {
    size_t nMatchings;
    Milliseconds time;
  };

  namespace blossom5 {
    template <typename Iter> blossom_result maximum_matching_size(Iter begin, Iter end, size_t N, size_t M) {
      auto start = std::chrono::high_resolution_clock::now();
      PerfectMatching pm(2 * N, 2 * M + N);

      for (auto iter = begin; iter != end; ++iter)
        pm.AddEdge(iter->x.id, iter->y.id, -1); // original edges
      for (auto iter = begin; iter != end; ++iter)
        pm.AddEdge(iter->x.id + N, iter->y.id + N, -1); // copies
      for (int i = 0; i < N; ++i)
        pm.AddEdge(i, i + N, 0); // edges between original nodes and their copies

      pm.options.verbose = false;
      pm.Solve();

      // get size of maximum matching size
      blossom_result result {};
      result.nMatchings = 0;
      for (int i = 0; i < M; i++) {
        if (pm.GetSolution(i)) result.nMatchings++;
      }

      auto tend = std::chrono::high_resolution_clock::now();
      result.time = duration(tend - start);

      return result;
    }

    namespace rnd {
      template <typename Iter> blossom_result maximum_matching_size(Iter begin, Iter end, size_t N, size_t M) {

        std::vector<size_t> perm(N);
        for(auto i = 0u; i < N; i++) {
          perm[i] = i;
        }

        std::srand(std::time(nullptr));
        std::random_shuffle(perm.begin(), perm.end(), [] (auto i) { return std::rand() % i; } );

        auto start = std::chrono::high_resolution_clock::now();
        PerfectMatching pm(2 * N, 2 * M + N);

        for (auto iter = begin; iter != end; ++iter)
          pm.AddEdge(perm[iter->x.id], perm[iter->y.id], -1); // original edges
        for (auto iter = begin; iter != end; ++iter)
          pm.AddEdge(perm[iter->x.id] + N, perm[iter->y.id] + N, -1); // copies
        for (int i = 0; i < N; ++i)
          pm.AddEdge(i, i + N, 0); // edges between original nodes and their copies

        pm.options.verbose = false;
        pm.Solve();

        blossom_result result {};
        result.nMatchings = 0;
        for (int i = 0; i < M; i++) {
          if (pm.GetSolution(i)) result.nMatchings++;
        }

        auto tend = std::chrono::high_resolution_clock::now();
        result.time = duration(tend - start);

        return result;
      }
    }
  } /* namespace blossom5 */
} /* namespace mmatching */
#endif /* __BLOSSOM5_HPP__ */
