#include <vector>
#include <queue>
#include <algorithm>
#include "graph.hpp"

namespace kernelizer {

  // Solve bipartite matching using Hopcroft-Karp algorithm
  struct bp {
    bp(graph::graph &g)
      : g(&g), n(g.N()), matchedRL(g.N(), -1), matched(g.N(), false) {}

    std::vector<int> build() {
      while (true) {
        bfs();
        available.assign(g->N(), true);
        bool increase = false;
        for (auto &u: *g) { 
          if (not matched[u.id] and dfs(u.id)) increase = true;
        }
        if (not increase) break;
      }

      return matchedRL;
    }

  private:
    graph::graph *g;
    int n;

    // Bipartite graph (U, W, E)
    // dist[u] (u in U) : minimum distance from an unmatched vertex
    // matchedRL[w] (w in W) : the other endpoint of matching
    std::vector<int> dist, matchedRL;

    // available[u] (u in U): true if not used in an augmenting path
    // matched[u] (u in U)  : true if an endpoint of matching
    std::vector<bool> available, matched;

    void bfs() {
      dist.assign(g->N(), -1);

      std::queue<int> que;
      for (auto &u : *g) {
        if (not matched[u.id]) {
          que.push(u.id);
          dist[u.id] = 0;
        }
      }

      while (!que.empty()) {
        int u1 = que.front();
        que.pop();
        for (auto &w : g->get_vertex(u1).neighbors) {
          int u2 = matchedRL[w.id];
          if (u2 >= 0 and dist[u2] < 0) {
            dist[u2] = dist[u1] + 1;
            que.push(u2);
          }
        }
      }
    }

    bool dfs(int u) {
      available[u] = false;
      for (auto &w: g->get_vertex(u).neighbors) {
        int uu = matchedRL[w.id];
        if (uu < 0 or (available[uu] and dist[uu] == dist[u] + 1 and dfs(uu))) {
          matchedRL[w.id] = u;
          matched[u] = true;
          return true;
        }
      }
      return false;
    }

  };
}