#ifndef INCLUDE_GRAPH_H_
#define INCLUDE_GRAPH_H_

#include <cstdio>
#include <vector>
#include <random>
// #include <numeric>


namespace graph {
  template<class T>
  struct list_iterator {
    list_iterator() {}
    list_iterator(T* e) :e(e) {}

    list_iterator& operator++() {
      e = e->next;
      return *this;
    }

    T& operator*() {
      return *e;  
    }
    const T& operator*() const {
      return *e;  
    }
    T* operator->() {
      return e;   
    }
    const T* operator->() const {
      return e;   
    }
    bool operator==(const list_iterator<T>& rhs) const  { return e == rhs.e; } 
    bool operator!=(const list_iterator<T>& rhs) const { return !(e == rhs.e); }
    private:
      T* e = nullptr;
  };

  struct vertex;
  struct edge;
  struct neighbor;

  struct adjacenylist {
    void add(neighbor* v);
    list_iterator<neighbor> begin();
    list_iterator<neighbor> end();
    void remove(neighbor& n);

    const list_iterator<neighbor> begin() const;
    const list_iterator<neighbor> end() const;
    neighbor* last() { return _last; }
    const neighbor* last() const { return _last; }
  private:
    neighbor* first = nullptr;
    neighbor* _last = nullptr;
  };

  struct neighbor {
    unsigned int id;
    edge *e;
    neighbor* prev = nullptr;
    neighbor* next = nullptr;
    neighbor* complement(); 
  };

  struct vertex {
    unsigned int id;
    unsigned int degree = 0;
    adjacenylist neighbors;
  };

  struct edge {
    neighbor x, y;
    edge *prev = nullptr;
    edge *next = nullptr;
  };

  struct edgelist {
    edgelist(size_t M) : list(M) {}

    void add(unsigned int source, unsigned int target) {
      if (counter == list.size()) throw "added too many edges";

      list[counter].x.id = source;
      list[counter].x.e = &list[counter];
      list[counter].y.id = target;
      list[counter].y.e = &list[counter];

      if (first) {
        list[counter].prev = _last;
        _last->next = &list[counter];
      } else {
        first = &list[counter];
      }

      _last = &list[counter];
      counter++;
      num_of_edges++;
    }

    void remove(edge& e) {
      if (e.prev == nullptr) {
        first = e.next;
      } else {
        e.prev->next = e.next;
      }

      if (e.next == nullptr) {
        _last = e.prev;
      } else {
        e.next->prev = e.prev;
      }
      num_of_edges--;
    }

    list_iterator<edge> begin() {
      return list_iterator<edge>(first);
    }
    list_iterator<edge> end() {
      return list_iterator<edge>();
    }

    const list_iterator<edge> begin() const {
      return list_iterator<edge>(first);
    }
    const list_iterator<edge> end() const {
      return list_iterator<edge>();
    }

    size_t M() const { return num_of_edges; }

    edge* last() { return _last; }

  private:
    size_t num_of_edges = 0;
    std::vector<edge> list;
    unsigned int counter = 0;
    edge* first = nullptr;
    edge* _last = nullptr;
  };

  struct graph {

    graph(unsigned int N, unsigned int M) :edges(M), vertices(N) {
      //set the ids
      for(size_t i = 0 ; i < N ; i++) {
        vertices[i].id = i;
      }
    }

    void add_edge(unsigned int source, unsigned int target) {
      edges.add(source, target);
      vertices[source].neighbors.add(&edges.last()->y);
      vertices[source].degree++;
      vertices[target].neighbors.add(&edges.last()->x);
      vertices[target].degree++;
    }

    void remove_edge(edge& e) {
      edges.remove(e);
      vertices[e.x.id].neighbors.remove(e.y);
      vertices[e.y.id].neighbors.remove(e.x);
      vertices[e.x.id].degree--;
      vertices[e.y.id].degree--;
    }

    vertex& get_vertex(unsigned int id) {
      return vertices[id];
    }
    const vertex& get_vertex(unsigned int id) const {
      return vertices[id];
    }

    const vertex& operator[](unsigned int id) const {
      return get_vertex(id);
    }
    vertex& operator[](unsigned int id) {
      return get_vertex(id);
    }

    const edgelist& edgeset() const { return edges; }
    unsigned int N() const { return vertices.size(); }

    std::vector<vertex>::iterator begin() { return vertices.begin(); }
    std::vector<vertex>::iterator end() { return vertices.end(); }

    void isolate_vertex(unsigned int id) {
      for(auto& v : vertices[id].neighbors) {
        remove_edge(*v.e);
      }
    }

  private:
    edgelist edges;
    std::vector<vertex> vertices;
  };

  template<class F>
  graph exclude_vertices(const graph &g, F f) {
    std::vector<unsigned int> names(g.N(), -1); 
    unsigned int c = 0;
    for(unsigned int i = 0; i < g.N(); i++) {
      if (!f(g[i])) {
        names[i] = c;
        c++;
      }
    }

    graph result(c, g.edgeset().M());
    for(auto& e : g.edgeset()) {
      if(names[e.x.id] != -1 && names[e.y.id] != -1)
        result.add_edge(names[e.x.id],names[e.y.id]);
    }
    return result;  
  }

  inline graph exclude_isolated_vertices(const graph& g) {
    return exclude_vertices(g, [] (const auto& v) { return v.degree == 0; });
  }

  inline graph generate_graph(size_t n, size_t p = 50) {
    std::mt19937 rng;
    rng.seed(std::random_device()());
    std::uniform_int_distribution<std::mt19937::result_type> dist(1, 100);
    graph g(n,n*n);
    for (auto i = 0u;i < n;i++) {
      for (auto j = i + 1; j < n; j++) {
        if (dist(rng) < p) g.add_edge(i, j);
      }
    }
    return g;
  }
 
}

#endif
