/*
*
* Author: Philipp Zschoche, https://zschoche.org
*
*/
#include <queue>
#include "graph.hpp"
#include "logger.hpp"

namespace kernelizer {

  /*
   * n, nPrime: number of vertices before (after) kernelization
   * m, mPrime: number of edges before (after) kernelization
   * nMatchings: number of matchings found during kernelization
   */
  struct k_degree_result {
    size_t n, m, nPrime, mPrime, nMatchings;
    unsigned int rule0, rule1, rule2;
    Milliseconds time;
  };

  struct ker_degree {

    ker_degree(graph::graph &g) : g(&g), collected(g.N(), false), mark(g.N(), false) {}

    k_degree_result kernelize() {
      k_degree_result result {};
      result.n = g->N();
      result.m = g->edgeset().M();
      result.nMatchings = 0;
      result.rule0 = 0;
      result.rule1 = 0;
      result.rule2 = 0;

      auto start = Clock::now();

      collect_vertices();  // find all vertices of degree one or two
      while (!V.empty()) {
        auto *v = V.front();
        V.pop();
        if (v->degree == 0) {
          result.rule0++;
        } else if (v->degree == 1) {
          result.nMatchings++;
          result.rule1++;
          remove_leaf(v);
        } else if (v->degree == 2) {
          result.nMatchings++;
          result.rule2++;
          remove_redex(v);
        }
      }

      auto end = Clock::now();
      result.time = duration(end - start);

      auto h = graph::exclude_isolated_vertices(*g);
      result.nPrime = h.N();
      result.mPrime = h.edgeset().M();

      return result;
    }

  private:
    graph::graph *g;
    std::vector<bool> collected, mark;
    std::queue<graph::vertex *> V;

    void remove_redex(graph::vertex *v) {
      auto& u = *v->neighbors.begin();
      auto& w = *u.next;
      auto& uu =g->get_vertex(u.id);
      auto& ww =g->get_vertex(w.id);
      g->remove_edge(*u.e);
      g->remove_edge(*w.e);
      if (uu.degree < ww.degree)
        merge(uu, ww);
      else 
        merge(ww, uu);
    }

    void merge(graph::vertex& u, graph::vertex& w) {
      for(auto& v : w.neighbors) {
        mark[v.id] = true;
      }
      auto iter = u.neighbors.begin();
      while (iter != u.neighbors.end()) {
        auto& v = (*iter);
        ++iter;
        if (v.id == w.id) {
          g->remove_edge(*v.e);
        } else {
          if (mark[v.id]) {
            //just remove this edge
            g->remove_edge(*v.e);
            //check degree
            auto& vv = g->get_vertex(v.id);
            collect(&vv);
          } else {
            //redirect edge
            auto* c = v.complement();
            c->id = w.id;
            u.neighbors.remove(v);
            u.degree--;
            v.next = nullptr;
            w.neighbors.add(&v);
            w.degree++;
          }
        }
      }

      collect(&w);

      // not optimal
      iter = w.neighbors.begin();
      while (iter != w.neighbors.end()) {
        mark[iter->id] = false;
        ++iter;
      }
    }

    void remove_leaf(graph::vertex *v) {
      auto &n = (*g)[v->neighbors.begin()->id];
      // isolate the deg 1 vertex
      g->remove_edge(*v->neighbors.begin()->e);

      // isolate the neighbor
      for (auto &w : n.neighbors) {
        g->remove_edge(*w.e);
        collect(&(*g)[w.id]); // there is a new deg 2 vertex.
      }
    }

    void collect_vertices() {
      for (auto &v : *g) {
        collect(&v);
      }
    }

    inline void collect(graph::vertex* v) {
      if (not (v->degree == 1 or v->degree == 2)) return;
      if (not collected[v->id]) V.push(v);
      collected[v->id] = true;
    }
  };

}

