#include <vector>
#include <queue>
#include <algorithm>
#include <cassert>
#include <chrono>
#include "bp.hpp"
#include "scc.hpp"
#include "logger.hpp"

namespace kernelizer {

  struct k_crown_result {
    size_t n, m, nMatchings, nOddCyles, nOddCycleVertices,
           lpSolution,  // LP
           n1, m1,      // part 1
           n2, m2,      // part 2
           d1, d2;      // degree counts
    Milliseconds lpTime, p1Time, p2Time, time;
  };

  struct kernelizer {

    // vertices on the left side have id in [0, n - 1]
    // vertices on the right side have id in [n, 2n - 1]
    kernelizer(graph::graph &g)
      : g(&g), n(g.N()), m_curr(0), matchedLR(g.N(), -1), fromS(2 * g.N(), false) {
        for (auto &u: g) {
          assert(u.id < n);
          m_curr += u.degree;
        }
        m_curr /= 2;
      }

    k_crown_result kernelize() {
      k_crown_result result = {};
      result.n = g->N();
      result.m = g->edgeset().M();
      result.nMatchings = 0;
      result.nOddCyles = 0;
      result.nOddCycleVertices = 0;

      // Solve LP
      auto lpStart = Clock::now();
      bp hopcroft(*g);
      matchedRL = hopcroft.build();
      build_matchedLR();
      auto lpEnd = Clock::now();
      result.lpTime = duration(lpEnd - lpStart);
      // LOG_PERFORMANCE("LP finished in %llu (ms)\n", duration(lpEnd - lpStart));

      for (int i = 0; i < n; ++i) {
        if (matchedRL[i] >= 0) {
          result.lpSolution++;
          // LOG_RESULTVV("lp matched: %d, %d\n", i + 1, matchedRL[i] + 1);
        }
      }
      // LOG_RESULT("LP solution: %d\n", lpSolution);

      // Part 1
      // Residual graph has an edge (source, v) iff u is not matched
      auto partOneStart = Clock::now();
      for (auto &v: *g) {
        if (matchedLR[v.id] < 0) {
          remove_flared_crown_dfs(v.id, 0);
        }
      }

      for (auto v: *g) {
        int vid = v.id;

        // x_v = 0 if l_v is reachable from s and r_v is not
        // note that for every v with x_v = 1, there exists u s.t.
        // u is a neighbor of v and (l_v, r_u) is matched in LP solution
        if (fromS[vid] and not fromS[vid + n]) {
          if (matchedLR[vid] >= 0) {
            result.nMatchings += 1;
            isolate_vertex(matchedLR[vid]);
            // LOG_RESULTV("Part 1. matched: %d, %d\n", vid + 1, matchedLR[vid] + 1);
          }
          isolate_vertex(vid);
        }
      }
      auto partOneEnd = Clock::now();
      result.p1Time = duration(partOneEnd - partOneStart);

      for (auto v: *g) {
        if (v.degree > 0) {
          result.n1++;
        }
      }
      result.m1 = m_curr;

      // Part 2
      auto partTwoStart = Clock::now();
      if (g->N() > 0) {
        scc kosaraju(*g, matchedRL, matchedLR);
        auto tOrder = kosaraju.build();
        auto nsccs = *std::max_element(tOrder.begin(), tOrder.end()) + 1;

        std::vector<std::vector<int> > sccs(nsccs);
        for (auto &v: *g) {
          sccs[tOrder[v.id]].push_back(v.id);  // left
          sccs[tOrder[v.id + n]].push_back(v.id + n);  // right
        }

        // whether i-th scc is removed
        std::vector<bool> removed(nsccs, false);

        for (int i = nsccs - 1; i >= 0; --i) {

          // Left side and right side are disjoint
          std::vector<bool> lVertices(n, false);
          for (auto v: sccs[i]) {
            if (v < n) lVertices[v] = true;
          }
          for (auto v: sccs[i]) {
            if (v >= n and lVertices[v - n]) {
              goto endloop;
            }
          }

          // no outgoing edge
          for (auto vlr: sccs[i]) {
            int v = vlr % n;
            int lr = vlr / n;
            if (not lr) {  // left side
              for (auto &nv: g->get_vertex(v).neighbors) {
                auto nvto = tOrder[nv.id + n];
                assert(nvto <= i);
                if (nvto > i and not removed[nvto]) goto endloop;
              }
            } else if (matchedLR[v] >= 0) {  // right side
              auto nvto = tOrder[matchedRL[v]];
              assert(nvto <= i);
              if (nvto > i and not removed[nvto]) goto endloop;
            }
          }

          removed[i] = true;

          for (auto vlr: sccs[i]) {
            int v = vlr % n;
            result.nMatchings += (vlr < n and g->get_vertex(v).degree > 0);
          }

          for (auto vlr: sccs[i]) {
            int v = vlr % n;
            isolate_vertex(v);
          }
endloop:;
        }
      }

      auto partTwoEnd = Clock::now();
      result.p2Time = duration(partTwoEnd - partTwoStart);
      result.time = result.p1Time + result.p2Time;

      result.d1 = 0;
      result.d2 = 0;
      for (auto v: *g) {
        if (v.degree > 0) {
          result.n2++;
        }
        result.d1 += v.degree == 1;
        result.d2 += v.degree == 2;
      }
      result.m2 = m_curr;

      // count number of odd cycles
      std::vector<bool> visited(n, false);
      for (int i = 0; i < n; ++i) {
        if (matchedRL[i] < 0 or visited[i]) continue;
        int curr = i;
        visited[i] = true;
        int len = 0;
        do {
          curr = matchedRL[curr];
          visited[curr] = true;
          len++;
        } while (curr != i);
        result.nOddCyles += len % 2;
        if (len % 2) {
          result.nOddCycleVertices += len;
        }
      }

      return result;
    }

  private:
    graph::graph *g;
    int n, m_curr;

    // reachable from root s
    std::vector<bool> fromS;

    std::vector<int> matchedRL, matchedLR;

    void build_matchedLR() {
      for (auto &v: *g) {
        int rv = v.id;
        int lv = matchedRL[rv];
        if (lv >= 0) {
          matchedLR[lv] = rv;
        }
      }
    }

    void remove_flared_crown_dfs(int v, int lr) {
      int vlr = v + n * lr;
      if (fromS[vlr]) return;
      fromS[vlr] = true;
      if (not lr) {  // left
        for (auto &vv: g->get_vertex(v).neighbors) {
          remove_flared_crown_dfs(vv.id, 1);
        }
      } else {
        if (matchedRL[v] >= 0) {
          remove_flared_crown_dfs(matchedRL[v], 0);
        }
      }
    }

    inline void isolate_vertex(int vid) {
      auto v = g->get_vertex(vid);
      m_curr -= v.degree;
      for (auto &nv: v.neighbors) {
        g->remove_edge(*nv.e);
      }
      matchedRL[vid] = matchedLR[vid] = -1;
    }

  };
}
