#include <cstdio>
#include <iostream>
#include <cassert>
#include <chrono>

#include "misc.hpp"
#include "degeneracy.hpp"
#include "ker_degree.hpp"
#include "kernelizer.hpp"
#include "blossom5.hpp"
#include "logger.hpp"

#include <boost/program_options.hpp>
#include "yaml-cpp/yaml.h"

namespace po = boost::program_options;

// For output serialization
YAML::Emitter out;

void serialize_misc_result(const analyzer::misc_result &);
void serialize_degeneracy_result(const analyzer::degeneracy_result &);
void serialize_k_degree_result(const kernelizer::k_degree_result &);
void serialize_k_crown_result(const kernelizer::k_crown_result &);
void serialize_blossom_result(const mmatching::blossom_result &);

graph::graph read_graph(const std::string &path) {
  FILE* fd = fopen(path.c_str(), "r");
  assert(fd);

  size_t N, M;
  fscanf(fd, "p %zu %zu\n", &N, &M);

  graph::graph g(N, M);

  size_t u, v;
  for (int i = 0; i < M; ++i) {
    fscanf(fd, "e %zu %zu\n", &u, &v);
    g.add_edge(u - 1, v - 1);
  }

  return g;
}

void write_graph(const std::string &path, graph::graph &g) {
  FILE* fd = fopen(path.c_str(), "w");
  assert(fd);

  fprintf(fd, "p %u %zu\n", g.N(), g.edgeset().M());

  for (auto &u: g) {
    for (auto &v: u.neighbors) {
      if (u.id < v.id) {
        fprintf(fd, "e %u %u\n", u.id + 1, v.id + 1);
      }
    }
  }
}

int main(int argc, char *argv[]) {
  std::string ifile, ofile;
  bool solve, skipMisc, skipKDeg, skipKCrown, debug;

  po::options_description desc("Available options");

  desc.add_options()
    ("help", "show help message")
    ("input,i", po::value(&ifile)->required(), "path of input graph file")
    ("output,o", po::value(&ofile), "write kernel result")
    ("blossom,s", po::bool_switch(&solve)->default_value(false), "just solve with blossom")
    ("without-misc", po::bool_switch(&skipMisc)->default_value(false), "skip misc (including degeneracy)")
    ("without-kdeg", po::bool_switch(&skipKDeg)->default_value(false), "no degree kernelization")
    ("without-kcrown", po::bool_switch(&skipKCrown)->default_value(false), "no crown kernelization")
    ("debug,d", po::bool_switch()->default_value(false), "show debug messages");

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);

  if (vm.count("help")) {
    std::cout << desc << std::endl;
    return 0;
  }

  if (solve) {
    skipMisc = true;
    skipKDeg = true;
    skipKCrown = true;
    ofile = "";
  }

  // LOG_RESULT("Load data from %s\n", ifile);
  auto g = read_graph(ifile);

    // Serialization begin
  out << YAML::BeginMap;
  out << YAML::Key << ifile;
  out << YAML::BeginMap;

  if (not skipMisc) {
    analyzer::misc misc_analyzer(g);
    auto misc_result = misc_analyzer.run();
    serialize_misc_result(misc_result);

    analyzer::degeneracy degeneracy_analyzer(g);
    auto degeneracy_result = degeneracy_analyzer.run();
    serialize_degeneracy_result(degeneracy_result);
  }

  kernelizer::k_degree_result kDegResult;
  if (not skipKDeg) {
    auto kerDegStart = Clock::now();
    kernelizer::ker_degree k_deg(g);
    kDegResult = k_deg.kernelize();
    g = graph::exclude_isolated_vertices(g);
    serialize_k_degree_result(kDegResult);
  }

  kernelizer::k_crown_result kCrownResult;
  if (not skipKCrown) {
    auto kernelizeStart = Clock::now();
    kernelizer::kernelizer k(g);
    kCrownResult = k.kernelize();
    // auto kernelizeEnd = Clock::now();
    // auto kernelizeDur = duration(kernelizeEnd - kernelizeStart);
    g = graph::exclude_isolated_vertices(g);
    serialize_k_crown_result(kCrownResult);
  }

  if (not ofile.empty()) {
    write_graph(ofile, g);
  }

  if (solve) {
    auto blossomResult = mmatching::blossom5::rnd::maximum_matching_size(g.edgeset().begin(), g.edgeset().end(), g.N(), g.edgeset().M());
    serialize_blossom_result(blossomResult);
  }

  out << YAML::EndMap;
  std::cout << out.c_str() << std::endl;
}

void serialize_misc_result(const analyzer::misc_result &result) {
  out << YAML::Key << "misc";
  out << YAML::BeginMap;
  out << YAML::Key << "n" << YAML::Value << result.n;
  out << YAML::Key << "m" << YAML::Value << result.m;
  out << YAML::Key << "degree_1" << YAML::Value << result.ns[1];
  out << YAML::Key << "degree_2" << YAML::Value << result.ns[2];
  out << YAML::Key << "degree_3" << YAML::Value << result.ns[3];
  out << YAML::Key << "degree_4" << YAML::Value << result.ns[4];
  out << YAML::Key << "degree_5" << YAML::Value << result.ns[5];
  out << YAML::Key << "max_degree" << YAML::Value << result.maxDegree;
  out << YAML::Key << "average_degree" << YAML::Value << result.averageDegree;
  out << YAML::Key << "oct_lower_bound" << YAML::Value << result.octBound;
  out << YAML::Key << "h_index" << YAML::Value << result.hindex;
  out << YAML::EndMap;
}

void serialize_degeneracy_result(const analyzer::degeneracy_result &result) {
  out << YAML::Key << "degeneracy";
  out << YAML::BeginMap;
  out << YAML::Key << "degeneracy" << YAML::Value << result.degeneracy;
  out << YAML::Key << "core_1" << YAML::Value << result.cores[1];
  out << YAML::Key << "core_2" << YAML::Value << result.cores[2];
  out << YAML::Key << "core_3" << YAML::Value << result.cores[3];
  out << YAML::Key << "core_4" << YAML::Value << result.cores[4];
  out << YAML::Key << "core_5" << YAML::Value << result.cores[5];
  out << YAML::EndMap;
}

void serialize_k_degree_result(const kernelizer::k_degree_result &result) {
  out << YAML::Key << "k_degree_result";
  out << YAML::BeginMap;
  out << YAML::Key << "n" << YAML::Value << result.n;
  out << YAML::Key << "m" << YAML::Value << result.m;
  out << YAML::Key << "n_prime" << YAML::Value << result.nPrime;
  out << YAML::Key << "m_prime" << YAML::Value << result.mPrime;
  out << YAML::Key << "matchings" << YAML::Value << result.nMatchings;
  out << YAML::Key << "rule_0" << YAML::Value << result.rule0;
  out << YAML::Key << "rule_1" << YAML::Value << result.rule1;
  out << YAML::Key << "rule_2" << YAML::Value << result.rule2;
  out << YAML::Key << "time" << YAML::Value << result.time;
  out << YAML::EndMap;
}

void serialize_k_crown_result(const kernelizer::k_crown_result &result) {
  out << YAML::Key << "k_crown_result";
  out << YAML::BeginMap;
  out << YAML::Key << "n" << YAML::Value << result.n;
  out << YAML::Key << "m" << YAML::Value << result.m;
  out << YAML::Key << "matchings" << YAML::Value << result.nMatchings;
  out << YAML::Key << "odd_cycles" << YAML::Value << result.nOddCyles;
  out << YAML::Key << "odd_cycle_vertices" << YAML::Value << result.nOddCycleVertices;
  out << YAML::Key << "lp_solution" << YAML::Value << result.lpSolution;
  out << YAML::Key << "n_part_1" << YAML::Value << result.n1;
  out << YAML::Key << "m_part_1" << YAML::Value << result.m1;
  out << YAML::Key << "n_part_2" << YAML::Value << result.n2;
  out << YAML::Key << "m_part_2" << YAML::Value << result.m2;
  out << YAML::Key << "degree_1_prime" << YAML::Value << result.d1;
  out << YAML::Key << "degree_2_prime" << YAML::Value << result.d2;
  out << YAML::Key << "lp_time" << YAML::Value << result.lpTime;
  out << YAML::Key << "time_part_1" << YAML::Value << result.p1Time;
  out << YAML::Key << "time_part_2" << YAML::Value << result.p2Time;
  out << YAML::Key << "time" << YAML::Value << result.time;
  out << YAML::EndMap;
}

void serialize_blossom_result(const mmatching::blossom_result &result) {
  out << YAML::Key << "blossom_result";
  out << YAML::BeginMap;
  out << YAML::Key << "mm_size" << YAML::Value << result.nMatchings;
  out << YAML::Key << "time" << YAML::Value << result.time;
  out << YAML::EndMap;
}
