#!/usr/bin/env python3

from argparse import ArgumentParser
import numpy as np
from pathlib import Path
import subprocess
import traceback
import yaml
import time
import csv


CWD = Path.cwd()
BUILD_DIR = CWD.parent / 'build'
EXE = BUILD_DIR / 'main'
BLOSSOM = BUILD_DIR / 'blossom'
SNAP_DATA_DIR = CWD.parent / 'data'
K_SNAP_DATA_DIR = CWD.parent / 'kernel'

DEFAULT_TIME_OUT = 3
DEFAULT_N_RUNS = 3


def run(command, timeout=False):
    if not timeout:
        output = subprocess.run(command, stdout=subprocess.PIPE, encoding='ascii')
    else:
        output = subprocess.run(command, stdout=subprocess.PIPE, encoding='ascii', timeout=args.timeout)
    output = output.stdout
    print(output)
    return output


def command_k(input_path):
    return (EXE, '--input', input_path, '--output', K_SNAP_DATA_DIR / input_path.name)


# solve without kernelization
def command_b(input_path):
    return (EXE, '--input', input_path, '--blossom')


def command_k1(input_path):
    return (EXE, '--input', input_path, '--without-misc', '--without-kcrown')


def command_k2(input_path):
    return (EXE, '--input', input_path, '--without-misc', '--without-kdeg')


def extract_value(dic):
    return next(iter(dic.values()))


def benchmark(input_path, k):
    path = input_path
    key = 'blossom_result'
    if k:
        path = K_SNAP_DATA_DIR / input_path.name
        key = 'k_' + key
    ts = []
    combined[graph_name][key] = { 'n_runs': 0 }
    for _ in range(args.n_runs):
        try:
            raw = run(command_b(path), timeout=True)
            blossom_result = extract_value(yaml.safe_load(raw))['blossom_result']
            mm_size = blossom_result['mm_size']
            ts.append(blossom_result['time'])
            combined[graph_name][key]['n_runs'] += 1
        except subprocess.TimeoutExpired:
            break
        except Exception as e:
            print(f'Unknown error encountered: {str(e)}')
            print(traceback.format_exc())
            continue

    if combined[graph_name][key]['n_runs'] > 0:
        combined[graph_name][key].update({
                    'mm_size': mm_size,
                    'time_average': np.mean(ts),
                    'time_median': np.median(ts),
                    'time_std': np.std(ts),
                    'time_min': np.min(ts),
                    'time_max': np.max(ts),
                    'time_lower_p': np.percentile(ts, 25),
                    'time_higher_p': np.percentile(ts, 75)
                })


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-t', '--timeout', type=int, default=DEFAULT_TIME_OUT, help='timeout in seconds')
    parser.add_argument('-n', '--n-runs', type=int, default=DEFAULT_N_RUNS)
    parser.add_argument('-o', '--output', default='results.csv')
    parser.add_argument('-s', '--size', type=int)
    parser.add_argument('--only', nargs='*')

    args = parser.parse_args()
    if args.timeout == DEFAULT_TIME_OUT or args.n_runs == DEFAULT_N_RUNS:
        print('[Caution] Using default time-out and n-runs values')

    # compile source combined
    run(('make', '-C', BUILD_DIR))

    paths = SNAP_DATA_DIR.glob('*.dimacs')
    if args.only:
        paths = [p for p in paths if any(p == Path(q).resolve() for q in args.only)]

    combined = {}
    for input_path in paths:
        if args.size is not None and input_path.stat().st_size > args.size:
            continue
        try:
            print('Start:', input_path)

            graph_name = input_path.stem
            combined[graph_name] = {}

            raw = run(command_k(input_path))
            results = yaml.safe_load(raw)
            contents = extract_value(results)
            print('Preprocessing altogether done')

            combined[graph_name] = contents

            raw = run(command_k1(input_path))
            combined[graph_name]['k1'] = extract_value(yaml.safe_load(raw))['k_degree_result']
            print('Preprocessing 1 done')

            raw = run(command_k2(input_path))
            combined[graph_name]['k2'] = extract_value(yaml.safe_load(raw))['k_crown_result']
            print('Preprocessing 2 done')

            print('Blossom start 1')
            benchmark(input_path, True)
            print('Blossom done 1')

            print('Blossom start 2')
            benchmark(input_path, False)
            print('Blossom done 2')

            if 'time_average' in combined[graph_name]['k_blossom_result'] and 'time_average' in combined[graph_name]['blossom_result']:
                combined[graph_name]['blossom_running_time'] = {
                    'with_k': combined[graph_name]['k_degree_result']['time'] + combined[graph_name]['k_crown_result']['lp_time'] + combined[graph_name]['k_crown_result']['time'] + combined[graph_name]['k_blossom_result']['time_average'],
                    'without_k': combined[graph_name]['blossom_result']['time_average']
                }

        except Exception as e:
            print(f'Error: {str(e)}')
            print(traceback.format_exc())
            continue

    d = []
    for i, (k, v) in enumerate(combined.items()):
        dic = {'id': i}
        dic['filename'] = k
        for kk, vv in v.items():
            for kkk, vvv in vv.items():
                dic[kk + '_' + kkk] = vvv
        d.append(dic)

    keys = ['id', 'filename',
            'misc_n', 'misc_m', 'misc_degree_1', 'misc_degree_2', 'misc_degree_3', 'misc_degree_4', 'misc_degree_5', 'misc_max_degree', 'misc_average_degree', 'misc_oct_lower_bound', 'misc_h_index',
            'degeneracy_degeneracy', 'degeneracy_core_1', 'degeneracy_core_2', 'degeneracy_core_3', 'degeneracy_core_4', 'degeneracy_core_5',
            'k_degree_result_n', 'k_degree_result_m', 'k_degree_result_n_prime', 'k_degree_result_m_prime', 'k_degree_result_matchings', 'k_degree_result_rule_0', 'k_degree_result_rule_1', 'k_degree_result_rule_2', 'k_degree_result_time',
            'k_crown_result_n', 'k_crown_result_m', 'k_crown_result_matchings','k_crown_result_odd_cycles', 'k_crown_result_odd_cycle_vertices', 'k_crown_result_lp_solution', 'k_crown_result_n_part_1', 'k_crown_result_m_part_1', 'k_crown_result_n_part_2', 'k_crown_result_m_part_2', 'k_crown_result_lp_time', 'k_crown_result_time_part_1', 'k_crown_result_time_part_2', 'k_crown_result_time', 'k_crown_result_degree_1_prime', 'k_crown_result_degree_2_prime',
            'k1_n', 'k1_m', 'k1_n_prime', 'k1_m_prime', 'k1_matchings', 'k1_rule_0', 'k1_rule_1', 'k1_rule_2', 'k1_time',
            'k2_n', 'k2_m', 'k2_matchings', 'k2_odd_cycles', 'k2_odd_cycle_vertices', 'k2_lp_solution', 'k2_n_part_1', 'k2_m_part_1', 'k2_n_part_2', 'k2_m_part_2', 'k2_lp_time', 'k2_time_part_1', 'k2_time_part_2', 'k2_time', 'k2_degree_1_prime', 'k2_degree_2_prime',
            'k_blossom_result_n_runs', 'k_blossom_result_mm_size', 'k_blossom_result_time_average', 'k_blossom_result_time_median', 'k_blossom_result_time_std', 'k_blossom_result_time_min', 'k_blossom_result_time_max', 'k_blossom_result_time_lower_p', 'k_blossom_result_time_higher_p',
            'blossom_result_n_runs', 'blossom_result_mm_size', 'blossom_result_time_average', 'blossom_result_time_median', 'blossom_result_time_std', 'blossom_result_time_min', 'blossom_result_time_max', 'blossom_result_time_lower_p', 'blossom_result_time_higher_p',
            'blossom_running_time_with_k', 'blossom_running_time_without_k']

    with open(args.output, 'w') as output_file:
        dict_writer = csv.DictWriter(output_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(d)
