#pragma once
#include "helpers.hpp"
#include <boost/program_options.hpp>
#include <fstream>
#include <iostream>
#include <numeric>
#include <random>
#include "logger.hpp"
#include <unordered_map>

namespace akt {
	
struct neighbor;
template<class T>
struct list_iterator {

	list_iterator() {}
	list_iterator(T* e) :e(e) {}

	list_iterator& operator++() {
		e = e->next;	
		return *this;
	}

	T& operator*() {
		return *e;	
	}
	const T& operator*() const {
		return *e;	
	}
	T* operator->() {
		return e;	
	}
	const T* operator->() const {
		return e;	
	}
	bool operator==(const list_iterator<T>& rhs) const  { return e == rhs.e; } 
	bool operator!=(const list_iterator<T>& rhs) const { return !(e == rhs.e); }
	private:
		T* e = nullptr;

};
struct adjacenylist {

	void add(neighbor* v);
	list_iterator<neighbor> begin();
	list_iterator<neighbor> end();
	void remove(neighbor& n);
		
	const list_iterator<neighbor> begin() const;
	const list_iterator<neighbor> end() const;
	neighbor* last() { return _last; }
	const neighbor* last() const { return _last; }
	private:
	neighbor* first = nullptr;
	neighbor* _last = nullptr;
};

struct vertex {
	unsigned int id;
	unsigned int degree = 0;
	adjacenylist neighbors;
};

struct edge;
struct neighbor {
	unsigned int id;
	edge *e;
	neighbor* prev = nullptr;
	neighbor* next = nullptr;
	neighbor* complement(); 

};

template <class OS> OS &operator<<(OS &os, const neighbor& n) {
	os << n.id;
	return os;
}


struct edge {
	neighbor x, y;
	edge *prev = nullptr;
	edge *next = nullptr;
	unsigned short weight = 1;
};

template <class OS> OS &operator<<(OS &os, const edge& e) {
	os << '{' << e.x.id << ',' << e.y.id << '}';
	return os;
}
struct edgelist {

	edgelist(size_t M) : list(M) {}

	void add(unsigned int source, unsigned int target, unsigned short weight) {

		if(counter == list.size()) {
			list.push_back(edge());	
		}

		list[counter].x.id = source;
		list[counter].x.e = &list[counter];
		list[counter].y.id = target;
		list[counter].y.e = &list[counter];
		list[counter].weight = weight;

		if(first) {
			list[counter].prev = _last;
			_last->next = &list[counter];
		} else {
			first = &list[counter];
		}

		_last = &list[counter];
		counter++;
		num_of_edges++;

	}

	void remove(edge& e) {
		if(e.prev == nullptr) {
			first = e.next;
		} else {
			e.prev->next = e.next;
		}

		if(e.next == nullptr) {
			_last = e.prev;
		} else {
			e.next->prev = e.prev;
		}
		num_of_edges--;
	}

	list_iterator<edge> begin() {
		return list_iterator<edge>(first);

	}
	list_iterator<edge> end() {
		return list_iterator<edge>();
	}

	const list_iterator<edge> begin() const {
		return list_iterator<edge>(first);

	}
	const list_iterator<edge> end() const {
		return list_iterator<edge>();
	}

	size_t M() const { return num_of_edges; }

	edge* last() { return _last; }

      private:
		size_t num_of_edges = 0;
	      std::deque<edge> list;
	      unsigned int counter = 0;
	      edge* first = nullptr;
	      edge* _last = nullptr;
};

template <class OS> OS &operator<<(OS &os, const edgelist& list) {
	for(auto& e : list) {
		os << e << ", ";
	}
	return os;
}
template <class OS> OS &operator<<(OS &os, const adjacenylist& list) {
	for(auto& e : list) {
		os << e << ", ";
	}
	return os;
}


struct graph {

	graph(unsigned int N, unsigned int M) :edges(M), vertices(N) {
		//set the ids
		for(size_t i = 0 ; i < N ; i++) {
			vertices[i].id = i;
		}
	}


	void add_edge(unsigned int source, unsigned int target, unsigned short weight = 1) {
		edges.add(source, target, weight);
		vertices[source].neighbors.add(&edges.last()->y);
		vertices[source].degree++;
		vertices[target].neighbors.add(&edges.last()->x);
		vertices[target].degree++;
	}

	void remove_edge(edge& e) {
		edges.remove(e);
		vertices[e.x.id].neighbors.remove(e.y);
		vertices[e.y.id].neighbors.remove(e.x);
		vertices[e.x.id].degree--;
		vertices[e.y.id].degree--;
	}
	template<class Iter>
	void remove_edges(Iter begin, Iter end) {
		while(begin != end) {
			auto& e = *begin;
			++begin;
			remove_edge(*e);
		}
	}

	vertex& get_vertex(unsigned int id) {
		return vertices[id];
	}
	const vertex& get_vertex(unsigned int id) const{
		return vertices[id];
	}
	

	const vertex& operator[](unsigned int id) const {
		return get_vertex(id);
	}
	vertex& operator[](unsigned int id) {
		return get_vertex(id);
	}

	const edgelist& edgeset() const { return edges; }
	unsigned int N() const { return vertices.size(); }

	std::vector<vertex>::iterator begin() { return vertices.begin(); }
	std::vector<vertex>::iterator end() { return vertices.end(); }

	void isolate_vertex(unsigned int id) {
		for(auto& v : vertices[id].neighbors) {
			remove_edge(*v.e);
		}
	}
	private:

	edgelist edges;
	std::vector<vertex> vertices;

};

template<class F>
graph exclude_vertices(const graph &g, F f) {
	std::vector<unsigned int> names(g.N(), -1);	
	unsigned int c = 0;
	for(unsigned int i = 0; i < g.N(); i++) {
		if(!f(g[i])) {
			names[i] = c;
			c++;
		}
	}

	graph result(c,g.edgeset().M());
	for(auto& e : g.edgeset()) {
		if(names[e.x.id] != -1 && names[e.y.id] != -1)
			result.add_edge(names[e.x.id],names[e.y.id],e.weight);
	}
	return result;	
}
inline graph exclude_isolated_vertices(const graph& g) {
	return exclude_vertices(g,[] (const auto& v) { return v.degree == 0; });
}
	


template<class IStream>
graph parse_snap(IStream&in, size_t N, size_t M) {

	auto begin = std::chrono::high_resolution_clock::now();
	graph g(N, M);
	std::string first, second;
	std::unordered_map<size_t, size_t> names;
	size_t counter = 0;

	while (std::getline(in, first) && first.size() > 0 && first[0] == '#')
		; // read comments
	auto i = first.find_first_of(" \t");
	second = first.substr(i + 1);
	first = first.substr(0, i);
	do {
		size_t v = std::atoi(first.c_str());
		size_t w = std::atoi(second.c_str());
		if(v < w) {
		auto iter = names.find(v);
		if (iter != names.end())
			v = iter->second;
		else {

			names[v] = counter;
			v = counter;
			counter++;
		}
		iter = names.find(w);
		if (iter != names.end())
			w = iter->second;
		else {

			names[w] = counter;
			w = counter;
			counter++;
		}
		 g.add_edge(v,w);
		}
	} while (in >> first >> second);
	auto end = std::chrono::high_resolution_clock::now();
	LOG(std::cout << std::setw(FW) << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << '\t';)
	return g;
}
template<class IStream>
graph parse_snap(IStream&in) {

	std::string first, second;
	std::unordered_map<size_t, size_t> names;
	size_t counter = 0;

	std::vector<std::set<size_t>> edges;


	while (std::getline(in, first) && first.size() > 0 && first[0] == '#')
		; // read comments
	auto i = first.find_first_of(" \t");
	second = first.substr(i + 1);
	first = first.substr(0, i);
	do {
		size_t v = std::atoi(first.c_str());
		size_t w = std::atoi(second.c_str());
		auto iter = names.find(v);
		if (iter != names.end())
			v = iter->second;
		else {

			names[v] = counter;
			v = counter;
			counter++;
			edges.push_back(std::set<size_t>());
		}
		iter = names.find(w);
		if (iter != names.end())
			w = iter->second;
		else {

			names[w] = counter;
			w = counter;
			counter++;
			edges.push_back(std::set<size_t>());
		}
		if(v > w) {
			std::swap(v,w);
		}

		if(v != w) {
			edges[v].insert(w);
		}
	} while (in >> first >> second);

	size_t m = 0;
	for(auto& i : edges) {
		m += i.size();
	}
	graph g(counter, m);
	for(int v = 0; v < edges.size(); v++) {
		for(auto w : edges[v]) {
			g.add_edge(v, w);
		}
	}
	return g;
}

inline graph generate_graph(size_t n, size_t p = 50) {
	std::mt19937 rng;
    	rng.seed(std::random_device()());
    	std::uniform_int_distribution<std::mt19937::result_type> dist(1,100);
	graph g(n,n*n);
	for(auto i = 0u;i < n;i++) {
		for(auto j = i+1;j < n;j++) {

			if(dist(rng) < p) {
				g.add_edge(i,j);	
			}
		}
	}
	return g; 
}

inline graph generate_weighted_graph(size_t n, size_t p = 50, unsigned short max_weight = 1000) {
	std::mt19937 rng;
    	rng.seed(std::random_device()());
    	std::uniform_int_distribution<std::mt19937::result_type> dist(1,100);
    	std::uniform_int_distribution<std::mt19937::result_type> distw(1,max_weight);
	graph g(n,n*n);
	for(auto i = 0u;i < n;i++) {
		for(auto j = i+1;j < n;j++) {

			if(dist(rng) < p) {
				g.add_edge(i,j, distw(rng));	
			}
		}
	}
	return exclude_isolated_vertices(g); 
}
template<class OStream>
OStream& to_dimacs(OStream& os, const graph& g) {
	os << "p " << g.N() << ' ' << g.edgeset().M() << std::endl;
	for(auto& e : g.edgeset()) {
		os << "e " << std::min(e.x.id, e.y.id)+1 << ' ' << std::max(e.x.id, e.y.id)+1 << std::endl;
	}
	return os;
}
template<class OStream>
OStream& to_dimacs_with_weights(OStream& os, const graph& g) {
	os << "p " << g.N() << ' ' << g.edgeset().M() << std::endl;
	for(auto& e : g.edgeset()) {
		os << "e " << std::min(e.x.id, e.y.id)+1 << ' ' << std::max(e.x.id, e.y.id)+1 << ' ' << e.weight << std::endl;
	}
	return os;
}





template<class IStream>
graph parse_dimacs(IStream& in) {

	auto begin = std::chrono::high_resolution_clock::now();

	std::string first, second;
	while (std::getline(in, first) && first.size() > 0 && first[0] != 'p'); // find start
	auto i = first.find_first_of(" \t", 2);
	second = first.substr(i + 1);
	first = first.substr(2, i);
	size_t N = std::atoi(first.c_str());
	size_t M = std::atoi(second.c_str());

	graph g(N, M);
	first.clear();
	while (std::getline(in, first)) {
		if(!first.empty() && first[0] == 'e') {
			//read edge
			auto i = first.find_first_of(" \t", 2);
			second = first.substr(i + 1);
			first = first.substr(2, i);
			size_t v = std::atoi(first.c_str());
			size_t w = std::atoi(second.c_str());
			g.add_edge(v-1,w-1);
		}
		first.clear();
	}
	auto end = std::chrono::high_resolution_clock::now();
	LOG(std::cout << std::setw(FW) << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << '\t';)
	return g;
}
template<class IStream>
graph parse_dimacs_with_weights(IStream& in) {

	auto begin = std::chrono::high_resolution_clock::now();

	std::string first, second, third;
	while (std::getline(in, first) && first.size() > 0 && first[0] != 'p'); // find start
	auto i = first.find_first_of(" \t", 2);
	second = first.substr(i + 1);
	first = first.substr(2, i);
	size_t N = std::atoi(first.c_str());
	size_t M = std::atoi(second.c_str());

	graph g(N, M);
	first.clear();
	while (std::getline(in, first)) {
		if(!first.empty() && first[0] == 'e') {
			//read edge
			auto i = first.find_first_of(" \t", 2);
			auto j = first.find_first_of(" \t", i+1);
			second = first.substr(i + 1,j - i);
			third = first.substr(j + 1);
			first = first.substr(2, i-2);
			size_t v = std::atoi(first.c_str());
			size_t w = std::atoi(second.c_str());
			size_t weight = std::atoi(third.c_str());
			g.add_edge(v-1,w-1, weight);
		}
		first.clear();
	}
	auto end = std::chrono::high_resolution_clock::now();
	LOG(std::cout << std::setw(FW) << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << '\t';)
	return g;
}
namespace rnd {
	
template<class IStream>
graph parse_dimacs(IStream& in) {
	std::srand ( unsigned ( std::time(0) ) );
	auto begin = std::chrono::high_resolution_clock::now();

	std::string first, second;
	while (std::getline(in, first) && first.size() > 0 && first[0] != 'p'); // find start
	auto i = first.find_first_of(" \t", 2);
	second = first.substr(i + 1);
	first = first.substr(2, i);
	size_t N = std::atoi(first.c_str());
	size_t M = std::atoi(second.c_str());

	std::vector<size_t> perm(N);
	for(auto i = 0u; i < N; i++) {
		perm[i] = i;
	}
	std::random_shuffle(perm.begin(), perm.end(), [] (auto i) { return std::rand()%i;} );

	graph g(N, M);
	first.clear();
	while (std::getline(in, first)) {
		if(!first.empty() && first[0] == 'e') {
			//read edge
			auto i = first.find_first_of(" \t", 2);
			second = first.substr(i + 1);
			first = first.substr(2, i);
			size_t v = std::atoi(first.c_str());
			size_t w = std::atoi(second.c_str());
			g.add_edge(perm[v-1],perm[w-1]);
		}
		first.clear();
	}
	auto end = std::chrono::high_resolution_clock::now();
	LOG(std::cout << std::setw(FW) << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << '\t';)
	return g;
}
template<class IStream>
graph parse_dimacs_with_weights(IStream& in) {
	std::srand ( unsigned ( std::time(0) ) );
	auto begin = std::chrono::high_resolution_clock::now();

	std::string first, second, third;
	while (std::getline(in, first) && first.size() > 0 && first[0] != 'p'); // find start
	auto i = first.find_first_of(" \t", 2);
	second = first.substr(i + 1);
	first = first.substr(2, i);
	size_t N = std::atoi(first.c_str());
	size_t M = std::atoi(second.c_str());

	std::vector<size_t> perm(N);
	for(auto i = 0u; i < N; i++) {
		perm[i] = i;
	}
	std::random_shuffle(perm.begin(), perm.end(), [] (auto i) { return std::rand()%i;} );

	graph g(N, M);
	first.clear();
	while (std::getline(in, first)) {
		if(!first.empty() && first[0] == 'e') {
			//read edge
			auto i = first.find_first_of(" \t", 2);
			auto j = first.find_first_of(" \t", i+1);
			second = first.substr(i + 1,j - i);
			third = first.substr(j + 1);
			first = first.substr(2, i-2);
			size_t v = std::atoi(first.c_str());
			size_t w = std::atoi(second.c_str());
			size_t weight = std::atoi(third.c_str());
			g.add_edge(perm[v-1],perm[w-1], weight);
		}
		first.clear();
	}
	auto end = std::chrono::high_resolution_clock::now();
	LOG(std::cout << std::setw(FW) << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << '\t';)
	return g;
}
} /* namespace rnd */

template<class IStream>
graph parse_dimacs_rnd_weights(IStream& in,unsigned short max_weight) {

	std::mt19937 rng;
    	rng.seed(std::random_device()());
    	std::uniform_int_distribution<std::mt19937::result_type> dist(1,max_weight); //
	auto begin = std::chrono::high_resolution_clock::now();

	std::string first, second;
	while (std::getline(in, first) && first.size() > 0 && first[0] != 'p'); // find start
	auto i = first.find_first_of(" \t", 2);
	second = first.substr(i + 1);
	first = first.substr(2, i);
	size_t N = std::atoi(first.c_str());
	size_t M = std::atoi(second.c_str());

	graph g(N, M);
	first.clear();
	while (std::getline(in, first)) {
		if(!first.empty() && first[0] == 'e') {
			//read edge
			auto i = first.find_first_of(" \t", 2);
			second = first.substr(i + 1);
			first = first.substr(2, i);
			size_t v = std::atoi(first.c_str());
			size_t w = std::atoi(second.c_str());
			g.add_edge(v-1, w-1, dist(rng));
		}
		first.clear();
	}
	auto end = std::chrono::high_resolution_clock::now();
	LOG(std::cout << std::setw(FW) << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << '\t';)
	return g;
}



}

namespace std {
template<class T>
struct iterator_traits<akt::list_iterator<T>> {
    typedef ptrdiff_t difference_type; //almost always ptrdiff_t
    typedef T value_type; //almost always T
    typedef T& reference; //almost always T& or const T&
    typedef T* pointer; //almost always T* or const T*
    typedef std::forward_iterator_tag iterator_category;  //usually std::forward_iterator_tag or similar
};
}

