/*
*
*	Author: Philipp Zschoche, https://zschoche.org
*
*/
#ifndef __W_KERNEL_HPP__
#define __W_KERNEL_HPP__

#include "graph.hpp"
#include <queue>
namespace mmatching {

inline void remove_zero_weight_edges(akt::graph &g) {
	auto iter = g.edgeset().begin();
	auto end = g.edgeset().end();
	while (iter != end) {
		if (iter->weight == 0) {
			auto &e = *iter;
			++iter;
			g.remove_edge(e);
		} else {
			++iter;
		}
	}
}
struct weighted_kernelizer {

	weighted_kernelizer(akt::graph *g) : g(g), leafs(g->N(), 0), R(g->N(), 0) {}

	size_t operator()() {
		std::queue<akt::vertex *> V2;
		size_t s;
		do {
			s = score;
			while (collect_leafs()) {
				while (!V.empty()) {
					auto *v = V.front();
					V.pop();
					if (v->degree == 1) {
						remove_leaf(v);
					}
				}
				update_all_edges();
			}
			for (auto &v : *g) {
				if (v.degree == 2) {
					V2.push(&v);
				}
			}
			while (!V2.empty()) {
				auto *v = V2.front();
				V2.pop();
				if (v->degree == 2) {
					//akt::to_dimacs_with_weights(std::cout, *g);
					remove_degree_2(v);
				}
			}
		} while (s != score);
		return score;
	}

      private:
	akt::graph *g;
	std::vector<size_t> leafs;
	std::vector<unsigned short> R;
	std::queue<akt::vertex *> V;
	size_t score = 0;

	void check_for_dead_edges(size_t id) {
		auto &v = (*g)[id];
		auto iter = v.neighbors.begin();
		auto end = v.neighbors.end();
		while (iter != end) {
			auto &n = *iter;
			auto weight = get_weight(*n.e);
			if (weight <= 0) {
				// std::cout << "dead-edge:" << n.e->x.id << ',' << n.e->y.id << ',' << get_weight(*n.e) << std::endl;
				// delete edge
				if ((*g)[n.id].degree == 2)
					V.push(&(*g)[n.id]);
				++iter;
				g->remove_edge(*n.e);
			} else {
				++iter;
			}
		}
		if (v.degree == 2)
			V.push(&v);
	}

	bool collect_leafs() {
		for (auto &v : *g) {
			if (v.degree == 1) {
				V.push(&v);
				leafs[v.neighbors.begin()->id]++;
			}
		}
		return !V.empty();
	}
	int get_weight(akt::edge &e) {

		int weight = e.weight;
		weight -= R[e.x.id];
		weight -= R[e.y.id];
		return weight;
	}
	void remove_leaf(akt::vertex *v) {
		auto n = v->neighbors.begin();
		auto *e = n->e;
		auto weight = get_weight(*e);
		//	std::cout << "weight:" << weight << std::endl;
		score += std::max<int>(weight, 0);
		//	std::cout << "score:" << score << std::endl;
		R[n->id] = std::max<int>(R[n->id], R[n->id] + weight);
		if (leafs[n->id] > 1) {
			leafs[n->id]--;
		}
		g->remove_edge(*e);
		if (leafs[n->id] == 0) {
			check_for_dead_edges(n->id);
		}
	}
	void update_all_edges() {
		auto iter = g->edgeset().begin();
		auto end = g->edgeset().end();
		while (iter != end) {
			auto weight = get_weight(*iter);
			if (weight <= 0) {
				auto &e = *iter;
				++iter;
				g->remove_edge(e);

			} else {
				iter->weight = weight;
				++iter;
			}
		}

		for (auto &i : leafs)
			i = 0;
		for (auto &i : R)
			i = 0;
	}

	std::pair<size_t, akt::edge *> get_next_vertex_on_path(size_t last, size_t cur) {
		auto &v = (*g)[cur];
		auto next = std::find_if(v.neighbors.begin(), v.neighbors.end(), [&](auto &w) { return w.id != last; });
		return std::make_pair(next->id, next->e);
	}
	template <class Iter> size_t maximum_matching(Iter begin, Iter end, unsigned short &last) {
		size_t x = 0;
		while (begin != end) {
			last = std::max(0, static_cast<int>((*begin)->weight) - last);
			x += last;
			++begin;
		}
		return x;
	}
	template <class Iter> size_t maximum_matching(Iter begin, Iter end) {
		unsigned short last = 0;
		return maximum_matching(begin, end, last);
	}

	void remove_cycle(size_t id, std::deque<akt::edge *> &E) {
		//for (auto *i : E) {
		//	std::cout << i->weight << std::endl;
		//}
		//std::cout << "--" << std::endl;
		auto x = (*E.begin())->weight;
		auto begin = ++E.begin();
		auto end = E.end();
		auto without_e = maximum_matching(begin, end);
		//std::cout << without_e << std::endl;
		end--;
		auto matched = maximum_matching(begin, end);
		begin++;
		auto with_e = maximum_matching(begin, end) + x;
		//std::cout << with_e << std::endl;

		auto w = std::max(without_e, with_e);
		score += w;
		w = w - matched;

		g->remove_edges(E.begin(), E.end());
		for (auto &n : (*g)[id].neighbors) {
			n.e->weight = std::max(static_cast<int>(n.e->weight) - static_cast<int>(w), 0);
		}
	}

	void remove_path(size_t s, size_t t, std::deque<akt::edge *> &E) {
		//for (auto *i : E) {
		//	std::cout << i->weight << std::endl;
		//}
		auto begin = E.begin();
		auto end = E.end();
		auto m = maximum_matching(begin, end);
		begin++;
		auto m_without_s = maximum_matching(begin, end);
		end--;
		auto m_without_t = maximum_matching(E.begin(), end);
		auto m_without_st = maximum_matching(begin, end);

		g->remove_edges(++E.begin(), E.end());

		E.front()->weight = m_without_t - m_without_st;
		size_t z;
		if (E.front()->x.id == s) {
			z = E.front()->y.id;
		} else {
			z = E.front()->x.id;
		}
		g->add_edge(z, t, m_without_s - m_without_st);

		auto iter = std::find_if((*g)[s].neighbors.begin(), (*g)[s].neighbors.end(), [&](auto &n) { return n.id == t; });
		if (iter == (*g)[s].neighbors.end()) {
			g->add_edge(s, t, m - m_without_st);
		} else {
			iter->e->weight = std::max<int>(iter->e->weight, m - m_without_st);
		}

		score += m_without_st;
	}

	void solve_cycle(const std::deque<akt::edge *> &E) {
		auto begin = ++E.begin();
		auto end = E.end();
		auto a = maximum_matching(begin, end);
		begin++;
		end--;
		auto b = maximum_matching(begin, end) + E.front()->weight;
		score += std::max<int>(b, a);
		g->remove_edges(E.begin(), E.end());
	}

	template <class Iter> void remove_leaf_path(size_t s, size_t t, Iter begin, Iter end) {
		unsigned short last = 0;
		auto m = maximum_matching(begin, end, last);
		score += m;
		g->remove_edges(begin, end);
		if (last > 0) {
			for (auto &n : (*g)[t].neighbors) {
				n.e->weight = std::max(0, static_cast<int>(n.e->weight) - static_cast<int>(last));
			}
		}
	}

	void remove_degree_2(akt::vertex *v) {
		std::deque<akt::edge *> E;
		// find start
		auto last = v->id;
		auto cur = v->neighbors.begin()->id;
		E.push_front(v->neighbors.begin()->e);
		while ((*g)[cur].degree == 2 && cur != v->id) {

			auto x = get_next_vertex_on_path(last, cur);
			//std::cout << "last=" << last;
			//std::cout << " cur=" << cur;
			//std::cout << " x.first=" << x.first << std::endl;
			E.push_front(x.second);
			last = cur;
			cur = x.first;
		}
		auto s = cur;
		if (s == v->id) {
			//std::cout << "cycle" << std::endl;
			solve_cycle(E);

		} else {
			// find end
			//std::cout << "find end" << std::endl;
			last = v->id;
			auto c = v->neighbors.begin();
			++c;
			cur = c->id;
			E.push_back(c->e);
			//std::cout << "last=" << last;
			//std::cout << " cur=" << cur << std::endl;
			while ((*g)[cur].degree == 2) {
				auto x = get_next_vertex_on_path(last, cur);
				//std::cout << "last=" << last;
				//std::cout << " cur=" << cur;
				//std::cout << " x.first=" << x.first << std::endl;
				E.push_back(x.second);
				last = cur;
				cur = x.first;
			}
			auto t = cur;
			//std::cout << "s=" << s << " t=" << t << std::endl;
			if ((*g)[t].degree == 1 && (*g)[s].degree == 1) {
				score += maximum_matching(E.begin(), E.end());
				g->remove_edges(E.begin(), E.end());
			} else if ((*g)[s].degree == 1) {
				remove_leaf_path(s, t, E.begin(), E.end());
			} else if ((*g)[t].degree == 1) {
				remove_leaf_path(t, s, E.rbegin(), E.rend());

			} else {
				if (s == t) {
					remove_cycle(s, E);
				} else {
					remove_path(s, t, E);
				}
			}
		}
	}
};

} /* namespace mmatching */

#endif /* __W_KERNEL_HPP__ */
